/***********************************************************************************

	Stock Clock Coordinator.c

	This file is based on Coordinator.c in the Microchip Stack for ZigBee version 1.0-1.35
	and you should be able to substitute the file directly in the demo application supplied
    with the stack

	All modified lines end in "//////"

	The stack may be downloaded from www.microchip.com

	You may contact the author Richard Hoptroff at rgh@flexipanel.com

************************************************************************************
/*
    Microchip ZigBee Stack

    Demo Coordinator

    This demonstration shows how a ZigBee coordinator can be set up.  This demo allows
    the PICDEM Z Demostration Board to act as either a "Switching Load Controller"
    (e.g. a light) or a "Switching Remote Control" (e.g. a switch) as defined by
    the Home Controls, Lighting profile.  It is designed to interact with a
    second PICDEM Z programmed with the Demo RFD project.

    To give the PICDEM Z "switch" capability, uncomment the I_AM_SWITCH definition
    below.  To give the PICDEM Z "light" capability, uncomment the I_AM_LIGHT
    definition below.  The PICDEM Z may have both capabilities enabled.  Be sure
    that the corresponding Demo RFD device is programmed with complementary
    capabilities.  NOTE - for simplicity, the ZigBee simple descriptors for this
    demonstration are fixed.

    If this node is configured as a "switch", it can discover the network address
    of the "light" using two methods.  If the USE_BINDINGS definition is
    uncommented below, then End Device Binding must be performed between the
    "switch" and the "light" before messages can be sent and received successfully.
    If USE_BINDINGS is commented out, then the node will default to the probable
    network address of the other node, and messages may be able to be sent
    immediately.  However, the node will also be capable of performing Device
    Discovery to discover the actual network address of the other node, in case
    the network was formed with alternate short address assignments.  NOTE: The
    USE_BINDINGS definition must be the same in both the RFD and the ZigBee
    Coordinator nodes.

    Switch functionality is as follows:
        RB4, I_AM_SWITCH defined, sends a "toggle" message to the other node's "light"
        RB4, I_AM_SWITCH not defined, no effect
        RB5, USE_BINDINGS defined, sends an End Device Bind request
        RB5, USE_BINDINGS undefined, sends a NWK_ADDR_req for the MAC address specified

    End Device Binding
    ------------------
    If the USE_BINDINGS definition is uncommented, the "switch" will send an
    APS indirect message to toggle the "light".  In order for the message to
    reach its final destination, a binding must be created between the "switch"
    and the "light".  To do this, press RB5 on one PICDEM Z, and then press RB5
    on the other PICDEM Z within 5 seconds.  A message will be displayed indicating
    if binding was successful or not.  Note that End Device Binding is a toggle
    function.  Performing the operation again will unbind the nodes, and messages
    will not reach their final destination.

    Device Discovery
    ----------------
    If the USE_BINDINGS definition is not uncommented, pressing RB5 will send a
    broadcast NWK_ADDR_req message.  The NWK_ADDR_req message contains the MAC
    address of the desired node.  Be sure this address matches the address
    contained in the other node's zigbee.def file.

 *********************************************************************
 * FileName:        Coordinator.c
 * Dependencies:
 * Processor:       PIC18F
 * Complier:        MCC18 v3.00 or higher
 * Company:         Microchip Technology, Inc.
 *
 * Software License Agreement
 *
 * The software supplied herewith by Microchip Technology Incorporated
 * (the Company) for its PICmicro Microcontroller is intended and
 * supplied to you, the Companys customer, for use solely and
 * exclusively on Microchip PICmicro Microcontroller products. The
 * software is owned by the Company and/or its supplier, and is
 * protected under applicable copyright laws. All rights are reserved.
 * Any use in violation of the foregoing restrictions may subject the
 * user to criminal sanctions under applicable laws, as well as to
 * civil liability for the breach of the terms and conditions of this
 * license.
 *
 * THIS SOFTWARE IS PROVIDED IN AN AS IS CONDITION. NO WARRANTIES,
 * WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED
 * TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. THE COMPANY SHALL NOT,
 * IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL OR
 * CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
 *
 *
 * Author               Date    Comment
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * DF/KO                01/09/06 Microchip ZigBee Stack v1.0-3.4.1
 ********************************************************************/

//******************************************************************************
// Header Files
//******************************************************************************

// Include the main ZigBee header file.
#include "zAPL.h"

// If you are going to send data to a terminal, include this file.
#include "console.h"

////// We use the nonvolatile memory routines
#include "zNVM.h"		//////

//******************************************************************************
// Configuration Bits
//******************************************************************************

////// Pixies can use internal oscillator or 16MHz oscillator if fitted.
////// PICDEM-Z has 4MHz oscillator fitted.  This, plus which I/O pins you use
////// are the only modification you need to consider when porting from PICDEM-Z to Pixie
#define OSC_INTERNAL	////// Internal 4MHz oscillator
//#define OSC_4MHz		////// PICDEM-Z 4MHz oscillator
//#define OSC_16MHz		////// Pixie 16MHz oscillator

#if defined(MCHP_C18) && defined(__18F4620)
    #pragma romdata CONFIG1H = 0x300001
#if defined(OSC_4MHz)									//////
    const rom unsigned char config1H = 0b00000110;      ////// HSPLL oscillator
#elif defined(OSC_16MHz)								//////
    const rom unsigned char config1H = 0b00000010;     	////// HS oscillator
#elif defined(OSC_INTERNAL)								//////
    const rom unsigned char config1H = 0b00001000;      ////// internal oscillator
#else													//////
#error "OSC_ not defined"								//////
#endif													//////	

    #pragma romdata CONFIG2L = 0x300002
    const rom unsigned char config2L = 0b00011111;      // Brown-out Reset Enabled in hardware @ 2.0V, PWRTEN disabled

    #pragma romdata CONFIG2H = 0x300003
    const rom unsigned char config2H = 0b00010010;      // HW WD disabled, 1:512 prescaler

    #pragma romdata CONFIG3H = 0x300005
    const rom unsigned char config3H = 0b10000000;      // PORTB digital on RESET

    #pragma romdata CONFIG4L = 0x300006
    const rom unsigned char config4L = 0b10000001;      // DEBUG disabled,
                                                        // XINST disabled
                                                        // LVP disabled
                                                        // STVREN enabled

    #pragma romdata

#elif defined(HITECH_C18) && defined(_18F4620)
    // Set configuration fuses for HITECH compiler.
    __CONFIG(1, 0x0600);    // HSPLL oscillator
    __CONFIG(2, 0x101F);    // PWRTEN disabled, BOR enabled @ 2.0V, HW WD disabled, 1:128 prescaler
    __CONFIG(3, 0x8000);    // PORTB digital on RESET
    __CONFIG(4, 0x0081);    // DEBUG disabled,
                            // XINST disabled
                            // LVP disabled
                            // STVREN enabled
#endif


//******************************************************************************
// Compilation Configuration
//******************************************************************************

#define USE_BINDINGS	//////
//#define I_AM_LIGHT	//////  The coordinator only transmits switch information
#define I_AM_SWITCH		//////  The coordinator only transmits switch information

////// The communications link uses the debug console Rx functions.  However, probably want to 
////// suppress the output to the PC since the RadioTicker app doesn't read it and it will just fill
////// up a buffer on the PC
#define	TX_DEBUG_OUTPUT	////// Comment out to turn off debug messages


extern ROM BYTE                 bindingTableUsageMap[BINDING_USAGE_MAP_SIZE];  //@@@@

//******************************************************************************
// Constants
//******************************************************************************

#define BIND_SWITCH                 RB5
#define LIGHT_SWITCH                RB4

#define BIND_INDICATION             LATA0
#define MESSAGE_INDICATION          LATA1

#define BIND_STATE_BOUND            0
#define BIND_STATE_TOGGLE           1
#define BIND_STATE_UNBOUND          1
//#define BIND_WAIT_DURATION          (5*ONE_SECOND)   		////// 
#define BIND_WAIT_DURATION          (60*ONE_SECOND)			////// Allow end device 60 seconds to bind from startup
															////// You should also modify CONFIG_ENDDEV_BIND_TIMEOUT in zdo.c 
															////// if you modify this value here

#define LIGHT_OFF                   0x00
#define LIGHT_ON                    0xFF
#define LIGHT_TOGGLE                0xF0


//******************************************************************************
// Application Variables
//******************************************************************************

ZIGBEE_PRIMITIVE    currentPrimitive;
SHORT_ADDR          destinationAddress;

static union
{
    struct
    {
        BYTE    bBroadcastSwitchToggled    : 1;
        BYTE    bLightSwitchToggled        : 1;
        BYTE    bTryingToBind              : 1;
        BYTE    bIsBound                   : 1;
        BYTE    bDestinationAddressKnown   : 1;
    } bits;
    BYTE Val;
} myStatusFlags;
#define STATUS_FLAGS_INIT       0x00
#define TOGGLE_BOUND_FLAG       0x08
#define bBindSwitchToggled      bBroadcastSwitchToggled

//////******************************************************************************

////// Definitions and declarations
#define IsBound	myStatusFlags.bits.bIsBound		////// Abbreviation to fit the listing in Circuit Cellar
#if !defined(TX_DEBUG_OUTPUT)
#  define 	ConsolePutROMString( x )	;		//////	Suppress debug output if required
#  define	PrintChar( x )	;
#endif

////// ROM data that gets modified must be in a reserved, aligned 0x40 region of memory 
////// to ensure NVMErase / NVMwrite functions do not modify executable code
#pragma romdata NV_MEM = 0xFD40				////// Align to 0x40 boundary before debug region
ROM	BOOL	NonvolatileBindFlag = FALSE;	////// Store bind state for next power-up
ROM	BYTE	NonvolatileFiller[0x3F];		////// fill up remainder of 0x40 section
#pragma romdata								//////

////// The following line is here only to generate an error if you have't read this.
////// In zMAC.h there is a timeout after which a message gets deleted if the sleepy
////// end device hasn't picked it up.  The default is 10 seconds, it sleeps for two
////// minutes, so the value has to be increased.  So find the definition for
////// MAC_PIB_macTransactionPersistenceTime in zMAC.h and change it to ten minutes as below
#define MAC_PIB_macTransactionPersistenceTime (ONE_SECOND * 10 * 60)


//******************************************************************************
//******************************************************************************
//******************************************************************************

void main(void)
{
	static BYTE	PointerPositionByte = 0;				////// The desired pointer position, multiples of 2 degrees, 0 - 179 please
	static BOOL bSendUpdate = FALSE;					////// TRUE if an update has been received and the end device should be informed

#ifdef OSC_INTERNAL			//////
	OSCCONbits.IRCF1 = 1;   ////// assuming <IDCF2:IDCF0>=100 on startup, changes to <IDCF2:IDCF0>=110 = 4MHz internal oscillator
	OSCTUNEbits.PLLEN = 1;  ////// PLL 4MHz -> 16MHz
#endif						//////
 
   	CLRWDT();
    ENABLE_WDT();

    currentPrimitive = NO_PRIMITIVE;

    // If you are going to send data to a terminal, initialize the UART.
	////// We will use the standard console routines to accept messages from RadioTicker.exe
	////// This also means we can test the firmware by tying characters from HyperTerminal
    ConsoleInit();

    // Initialize the ZigBee Stack.
    ZigBeeInit();

    // *************************************************************************
    // Perform any other initialization here
    // *************************************************************************

    myStatusFlags.Val = STATUS_FLAGS_INIT;
    destinationAddress.Val = 0x796F;    // Default to first RFD

	////// Get bind flag from nonvolatile memory
	IsBound = NonvolatileBindFlag;	//////

	////// if not bound, bind immediately on startup.  Note this make the implementation very easy but to rebind you need
	////// to reload the firmware on both tyhe coordinator and the end device.  In a commercial application, you would 
    ////// want to use a bind button.  If you look at the data sheet for Pixie Switcher (on www.flexipanel.com) you'll 
    ////// get an idea of how to implement things like this
	if (!IsBound) myStatusFlags.bits.bBindSwitchToggled = TRUE;  //////

    BIND_INDICATION = !myStatusFlags.bits.bIsBound;
    LATA4 = 0;

////// As you can see from the next few lines, this template code from Microchip is actually
////// their beta version, but the release version will be vailable by the time the article is
////// in press, so you would be wise to download the latest version of the Microchip source
////// and modify that.
    ConsolePutROMString( (ROM char *)"\r\n\r\n\r\n********************************\r\n" );
    ConsolePutROMString( (ROM char *)"Microchip ZigBee(TM) Stack - pre-v1.0-4.0!\r\n\r\n" );
    ConsolePutROMString( (ROM char *)"ZigBee Coordinator - Stock Clock Version\r\n\r\n" );		//////

// make things easy by clearing bind state and neighbor tables
{//@@@@
NWKClearNeighborTable();
ClearBindingTable();  // APSInit does this - why?
IsBound = 0;
myStatusFlags.bits.bBindSwitchToggled = TRUE;
}
   while (1)
    {
        CLRWDT();
        ZigBeeTasks( &currentPrimitive );

        switch (currentPrimitive)
        {
            case NLME_NETWORK_FORMATION_confirm:
                if (!params.NLME_NETWORK_FORMATION_confirm.Status)
                {
                    ConsolePutROMString( (ROM char *)"PAN " );
                    PrintChar( macPIB.macPANId.byte.MSB );
                    PrintChar( macPIB.macPANId.byte.LSB );
                    ConsolePutROMString( (ROM char *)" started successfully.\r\n" );
                    params.NLME_PERMIT_JOINING_request.PermitDuration = 0xFF;
                    currentPrimitive = NLME_PERMIT_JOINING_request;
                }
                else
                {
                    PrintChar( params.NLME_NETWORK_FORMATION_confirm.Status );
                    ConsolePutROMString( (ROM char *)" Error forming network.  Trying again...\r\n" );
                    currentPrimitive = NO_PRIMITIVE;
                }
                break;

            case NLME_PERMIT_JOINING_confirm:
                if (!params.NLME_PERMIT_JOINING_confirm.Status)
                {
                    ConsolePutROMString( (ROM char *)"Joining permitted.\r\n" );
                    currentPrimitive = NO_PRIMITIVE;
                }
                else
                {
                    PrintChar( params.NLME_PERMIT_JOINING_confirm.Status );
                    ConsolePutROMString( (ROM char *)" Join permission unsuccessful. We cannot allow joins.\r\n" );
                    currentPrimitive = NO_PRIMITIVE;
                }
                break;

            case APSDE_DATA_indication:
                {
                    WORD_VAL    attributeId;
                    BYTE        command;
                    BYTE        data;
                    //BYTE        dataType;
                    BYTE        frameHeader;
                    BYTE        sequenceNumber;
                    BYTE        transaction;
                    BYTE        transByte;

                    currentPrimitive = NO_PRIMITIVE;
                    frameHeader = APLGet();

                    switch (params.APSDE_DATA_indication.DstEndpoint)
                    {
                        case EP_ZDO:
                            #define dataLength command

                            // Put code here to handle any ZDO responses that we requested
                            if ((frameHeader & APL_FRAME_TYPE_MASK) == APL_FRAME_TYPE_MSG)
                            {
                                frameHeader &= APL_FRAME_COUNT_MASK;
                                for (transaction=0; transaction<frameHeader; transaction++)
                                {
                                    sequenceNumber          = APLGet();
                                    dataLength              = APLGet();
                                    transByte               = 1;    // Account for status byte

                                    switch( params.APSDE_DATA_indication.ClusterId )
                                    {

                                        // ********************************************************
                                        // Put a case here to handle each ZDO response that we requested.
                                        // ********************************************************

                                        #ifndef USE_BINDINGS
                                        case NWK_ADDR_rsp:
                                            if (APLGet() == SUCCESS)
                                            {
                                                ConsolePutROMString( (ROM char *)"Receiving NWK_ADDR_rsp.\r\n" );

                                                // Skip over the IEEE address of the responder.
                                                for (data=0; data<8; data++)
                                                {
                                                    APLGet();
                                                    transByte++;
                                                }
                                                destinationAddress.byte.LSB = APLGet();
                                                destinationAddress.byte.MSB = APLGet();
                                                transByte += 2;
                                                myStatusFlags.bits.bDestinationAddressKnown = 1;
                                            }
                                            break;
                                        #endif

                                        #ifdef USE_BINDINGS
                                        case END_DEVICE_BIND_rsp:
                                            switch( APLGet() )
                                            {
                                                case SUCCESS:
                                                    ConsolePutROMString( (ROM char *)"End device bind/unbind successful!\r\n" );
                                                    myStatusFlags.bits.bIsBound ^= 1;
                                                    BIND_INDICATION = !myStatusFlags.bits.bIsBound;

													{	////// Store new bind state in nonvolatile memory
														BYTE flag = myStatusFlags.bits.bIsBound;							//////
														NVMWrite((NVM_ADDR*)&NonvolatileBindFlag, &flag, sizeof(BYTE)); 	//////
													}	//////

                                                    break;
                                                case ZDO_NOT_SUPPORTED:
                                                    ConsolePutROMString( (ROM char *)"End device bind/unbind not supported.\r\n" );
                                                    break;
                                                case END_DEVICE_BIND_TIMEOUT:
                                                    ConsolePutROMString( (ROM char *)"End device bind/unbind time out.\r\n" );
                                                    break;
                                                case END_DEVICE_BIND_NO_MATCH:
                                                    ConsolePutROMString( (ROM char *)"End device bind/unbind failed - no match.\r\n" );
                                                    break;
                                                default:
                                                    ConsolePutROMString( (ROM char *)"End device bind/unbind invalid response.\r\n" );
                                                    break;
                                            }
                                            myStatusFlags.bits.bTryingToBind = 0;
                                            break;
                                        #endif

                                        default:
                                            break;
                                    }

                                    // Read out the rest of the MSG in case there is another transaction.
                                    for (; transByte<dataLength; transByte++)
                                    {
                                        APLGet();
                                    }
                                }
                            }
                            #undef dataLength
                            break;

                        // ************************************************************************
                        // Place a case for each user defined endpoint.
                        // ************************************************************************

                        case EP_LIGHT:
                            if ((frameHeader & APL_FRAME_TYPE_MASK) == APL_FRAME_TYPE_KVP)
                            {
                                frameHeader &= APL_FRAME_COUNT_MASK;
                                for (transaction=0; transaction<frameHeader; transaction++)
                                {
                                    sequenceNumber          = APLGet();
                                    command                 = APLGet();
                                    attributeId.byte.LSB    = APLGet();
                                    attributeId.byte.MSB    = APLGet();

                                    //dataType = command & APL_FRAME_DATA_TYPE_MASK;
                                    command &= APL_FRAME_COMMAND_MASK;

                                    if ((params.APSDE_DATA_indication.ClusterId == OnOffSRC_CLUSTER) &&
                                        (attributeId.Val == OnOffSRC_OnOff))
                                    {
                                        if ((command == APL_FRAME_COMMAND_SET) ||
                                            (command == APL_FRAME_COMMAND_SETACK))
                                        {
                                            // Prepare a response in case it is needed.
                                            TxBuffer[TxData++] = APL_FRAME_TYPE_KVP | 1;    // KVP, 1 transaction
                                            TxBuffer[TxData++] = sequenceNumber;
 //                                           TxBuffer[TxData++] = APL_FRAME_COMMAND_SET_RES | TRANS_UINT8 ;			////// Error in Microchip demo application
                                          	TxBuffer[TxData++] = APL_FRAME_COMMAND_SET_RES | (TRANS_UINT8 << 4);		////// TRANS_UINT8 must be left-shifted by 4
                                            TxBuffer[TxData++] = attributeId.byte.LSB;
                                            TxBuffer[TxData++] = attributeId.byte.MSB;

                                            // Data type for this attibute must be TRANS_UINT8
                                            data = APLGet();
                                            switch (data)
                                            {
                                                case LIGHT_OFF:
                                                    ConsolePutROMString( (ROM char *)"Turning light off.\r\n" );
                                                    MESSAGE_INDICATION = 0;
                                                    TxBuffer[TxData++] = SUCCESS;
                                                    break;
                                                case LIGHT_ON:
                                                    ConsolePutROMString( (ROM char *)"Turning light on.\r\n" );
                                                    MESSAGE_INDICATION = 1;
                                                    TxBuffer[TxData++] = SUCCESS;
                                                    break;
                                                case LIGHT_TOGGLE:
                                                    ConsolePutROMString( (ROM char *)"Toggling light.\r\n" );
                                                    MESSAGE_INDICATION ^= 1;
                                                    TxBuffer[TxData++] = SUCCESS;
                                                    break;
                                                default:
                                                    PrintChar( data );
                                                    ConsolePutROMString( (ROM char *)" Invalid light message.\r\n" );
                                                    TxBuffer[TxData++] = KVP_INVALID_ATTRIBUTE_DATA;
                                                    break;
                                            }
                                        }
                                        if (command == APL_FRAME_COMMAND_SETACK)
                                        {
                                            // Send back an application level acknowledge.
                                            ZigBeeBlockTx();

                                            // Take care here that parameters are not overwritten before they are used.
                                            // We can use the data byte as a temporary variable.
                                            params.APSDE_DATA_request.DstAddrMode = params.APSDE_DATA_indication.SrcAddrMode;
                                            params.APSDE_DATA_request.DstEndpoint = params.APSDE_DATA_indication.SrcEndpoint;
                                            params.APSDE_DATA_request.DstAddress.ShortAddr = params.APSDE_DATA_indication.SrcAddress.ShortAddr;

                                            //params.APSDE_DATA_request.asduLength; TxData
                                            //params.APSDE_DATA_request.ProfileId; unchanged
                                            params.APSDE_DATA_request.RadiusCounter = DEFAULT_RADIUS;
                                            params.APSDE_DATA_request.DiscoverRoute = ROUTE_DISCOVERY_ENABLE;
                                            params.APSDE_DATA_request.TxOptions.Val = 0;
                                            params.APSDE_DATA_request.SrcEndpoint = EP_LIGHT;
                                            //params.APSDE_DATA_request.ClusterId; unchanged

                                            currentPrimitive = APSDE_DATA_request;
                                        }
                                        else
                                        {
                                            // We are not sending an acknowledge, so reset the transmit message pointer.
                                            TxData = TX_DATA_START;
                                        }
                                    }
                                    // TODO read to the end of the transaction.
                                } // each transaction
                            } // frame type
                            break;
                        default:
                            // If the command type was something that requested an acknowledge, we could send back
                            // KVP_INVALID_ENDPOINT here.
                            break;
                    }
                    APLDiscardRx();
                }
                break;

            case APSDE_DATA_confirm:
                if (params.APSDE_DATA_confirm.Status)
                {
                    ConsolePutROMString( (ROM char *)"Error " );
                    PrintChar( params.APSDE_DATA_confirm.Status );
                    ConsolePutROMString( (ROM char *)" sending message.\r\n" );
                }
                else
                {
                    ConsolePutROMString( (ROM char *)" Message sent successfully.\r\n" );
                }
                currentPrimitive = NO_PRIMITIVE;
                break;

            case NO_PRIMITIVE:
                if (!ZigBeeStatus.flags.bits.bNetworkFormed)
                {
                    if (!ZigBeeStatus.flags.bits.bTryingToFormNetwork)
                    {
                        ConsolePutROMString( (ROM char *)"Trying to start network...\r\n" );
                        params.NLME_NETWORK_FORMATION_request.ScanDuration          = 8;
                        params.NLME_NETWORK_FORMATION_request.ScanChannels.Val      = ALLOWED_CHANNELS;
                        params.NLME_NETWORK_FORMATION_request.PANId.Val             = 0xFFFF;
                        params.NLME_NETWORK_FORMATION_request.BeaconOrder           = MY_BEACON_ORDER;
                        params.NLME_NETWORK_FORMATION_request.SuperframeOrder       = MY_SUPERFRAME_ORDER;
                        params.NLME_NETWORK_FORMATION_request.BatteryLifeExtension  = MY_BATTERY_LIFE_EXTENSION;
                        currentPrimitive = NLME_NETWORK_FORMATION_request;
                    }
                }
                else
                {
                    if (ZigBeeReady())
                    {

                        // ************************************************************************
                        // Place all processes that can send messages here.  Be sure to call
                        // ZigBeeBlockTx() when currentPrimitive is set to APSDE_DATA_request.
                        // ************************************************************************

                        #ifdef I_AM_SWITCH
//////                        if ( myStatusFlags.bits.bLightSwitchToggled )
                        if ( IsBound && bSendUpdate )		////// Send update to other node
                        {
                            // Send a light toggle message to the other node.
//////                            myStatusFlags.bits.bLightSwitchToggled = FALSE;
							bSendUpdate = FALSE; //////

				            ConsolePutROMString( (ROM char *)" Forwarding data = " ); 	//////
							PrintChar( PointerPositionByte ); 							//////
				            ConsolePutROMString( (ROM char *)"\r\n" ); 					//////

                            TxBuffer[TxData++] = APL_FRAME_TYPE_KVP | 1;    // KVP, 1 transaction
                            TxBuffer[TxData++] = APLGetTransId();
//                            TxBuffer[TxData++] = APL_FRAME_COMMAND_SET | TRANS_UINT8;				////// Error in Microchip demo application
                            TxBuffer[TxData++] = APL_FRAME_COMMAND_SET | (TRANS_UINT8 << 4);		////// TRANS_UINT8 must be left-shifted by 4
                            TxBuffer[TxData++] = OnOffSRC_OnOff & 0xFF;         // Attribute ID LSB
                            TxBuffer[TxData++] = (OnOffSRC_OnOff >> 8) & 0xFF;  // Attribute ID MSB
//////                            TxBuffer[TxData++] = LIGHT_TOGGLE;
                            TxBuffer[TxData++] = PointerPositionByte;			// "Fake" light message			

                            #ifdef USE_BINDINGS
                                // We are sending indirect
                                params.APSDE_DATA_request.DstAddrMode = APS_ADDRESS_NOT_PRESENT;
                            #else
                                // We are sending direct
                                params.APSDE_DATA_request.DstAddrMode = APS_ADDRESS_16_BIT;
                                params.APSDE_DATA_request.DstEndpoint = EP_LIGHT;
                                params.APSDE_DATA_request.DstAddress.ShortAddr = destinationAddress;
                            #endif

                            //params.APSDE_DATA_request.asduLength; TxData
                            params.APSDE_DATA_request.ProfileId.Val = MY_PROFILE_ID;
                            params.APSDE_DATA_request.RadiusCounter = DEFAULT_RADIUS;
                            params.APSDE_DATA_request.DiscoverRoute = ROUTE_DISCOVERY_ENABLE;
                            params.APSDE_DATA_request.TxOptions.Val = 0;
                            params.APSDE_DATA_request.SrcEndpoint = EP_LIGHT;
                            params.APSDE_DATA_request.ClusterId = OnOffSRC_CLUSTER;

                            ConsolePutROMString( (ROM char *)" Trying to send light switch message.\r\n" );

                            currentPrimitive = APSDE_DATA_request;
                        }
                        else
                        #endif
                        #ifdef USE_BINDINGS
                        if (myStatusFlags.bits.bBindSwitchToggled)
                        {
                            myStatusFlags.bits.bBindSwitchToggled = FALSE;
                            if (myStatusFlags.bits.bTryingToBind)
                            {
                                ConsolePutROMString( (ROM char *)" End Device Binding already in progress.\r\n" );
                            }
                            else
                            {
                                myStatusFlags.bits.bTryingToBind = 1;

                                // Send down an end device bind primitive.
                                ConsolePutROMString( (ROM char *)" Trying to perform end device binding.\r\n" );
                                currentPrimitive = ZDO_END_DEVICE_BIND_req;

                                params.ZDO_END_DEVICE_BIND_req.ProfileID.Val = MY_PROFILE_ID;
                                #ifdef I_AM_LIGHT
                                    params.ZDO_END_DEVICE_BIND_req.NumInClusters = 1;
                                    if ((params.ZDO_END_DEVICE_BIND_req.InClusterList=SRAMalloc(1)) != NULL)
                                    {
                                        *params.ZDO_END_DEVICE_BIND_req.InClusterList = OnOffSRC_CLUSTER;
                                    }
                                    else
                                    {
                                        myStatusFlags.bits.bTryingToBind = 0;
                                        ConsolePutROMString( (ROM char *)" Could not send down ZDO_END_DEVICE_BIND_req.\r\n" );
                                        currentPrimitive = NO_PRIMITIVE;
                                    }
                                #else
                                    params.ZDO_END_DEVICE_BIND_req.NumInClusters = 0;
                                    params.ZDO_END_DEVICE_BIND_req.InClusterList = NULL;
                                #endif
                                #ifdef I_AM_SWITCH
                                    params.ZDO_END_DEVICE_BIND_req.NumOutClusters = 1;
                                    if ((params.ZDO_END_DEVICE_BIND_req.OutClusterList=SRAMalloc(1)) != NULL)
                                    {
                                        *params.ZDO_END_DEVICE_BIND_req.OutClusterList = OnOffSRC_CLUSTER;
                                    }
                                    else
                                    {
                                        myStatusFlags.bits.bTryingToBind = 0;
                                        ConsolePutROMString( (ROM char *)" Could not send down ZDO_END_DEVICE_BIND_req.\r\n" );
                                        currentPrimitive = NO_PRIMITIVE;
                                    }
                                #else
                                    params.ZDO_END_DEVICE_BIND_req.NumOutClusters = 0;
                                    params.ZDO_END_DEVICE_BIND_req.OutClusterList = NULL;
                                #endif
                                params.ZDO_END_DEVICE_BIND_req.LocalCoordinator = macPIB.macShortAddress;
                                params.ZDO_END_DEVICE_BIND_req.endpoint = EP_LIGHT;
                            }
                        }
                        #else
                        if (myStatusFlags.bits.bBroadcastSwitchToggled)
                        {
                            myStatusFlags.bits.bBroadcastSwitchToggled = FALSE;

                            // Send NWK_ADDR_req message

                            TxBuffer[TxData++] = APL_FRAME_TYPE_MSG | 1;    // KVP, 1 transaction
                            TxBuffer[TxData++] = APLGetTransId();
                            TxBuffer[TxData++] = 10; // Transaction Length

                            // IEEEAddr
                            TxBuffer[TxData++] = 0x65;
                            TxBuffer[TxData++] = 0x00;
                            TxBuffer[TxData++] = 0x00;
                            TxBuffer[TxData++] = 0x00;
                            TxBuffer[TxData++] = 0x00;
                            TxBuffer[TxData++] = 0xa3;
                            TxBuffer[TxData++] = 0x04;
                            TxBuffer[TxData++] = 0x00;

                            // RequestType
                            TxBuffer[TxData++] = 0x00;

                            // StartIndex
                            TxBuffer[TxData++] = 0x00;

                            params.APSDE_DATA_request.DstAddrMode = APS_ADDRESS_16_BIT;
                            params.APSDE_DATA_request.DstEndpoint = EP_ZDO;
                            params.APSDE_DATA_request.DstAddress.ShortAddr.Val = 0xFFFF;

                            //params.APSDE_DATA_request.asduLength; TxData
                            params.APSDE_DATA_request.ProfileId.Val = MY_PROFILE_ID;
                            params.APSDE_DATA_request.RadiusCounter = DEFAULT_RADIUS;
                            params.APSDE_DATA_request.DiscoverRoute = ROUTE_DISCOVERY_ENABLE;
                            params.APSDE_DATA_request.TxOptions.Val = 0;
                            params.APSDE_DATA_request.SrcEndpoint = EP_ZDO;
                            params.APSDE_DATA_request.ClusterId = NWK_ADDR_req;

                            ConsolePutROMString( (ROM char *)" Trying to send NWK_ADDR_req.\r\n" );

                            currentPrimitive = APSDE_DATA_request;
                        }
                        #endif

                        // Re-enable interrupts.
                        RBIE = 1;
                    }
                }
                break;

            default:
                PrintChar( currentPrimitive );
                ConsolePutROMString( (ROM char *)" Unhandled primitive.\r\n" );
                currentPrimitive = NO_PRIMITIVE;
                break;
        }

        // *********************************************************************
        // Place any non-ZigBee related processing here.  Be sure that the code
        // will loop back and execute ZigBeeTasks() in a timely manner.
        // *********************************************************************

		////// If a message is received from the radio ticker, store it and raise a flag 
		////// to indicate that a message must be sent to end device
		if (ConsoleIsGetReady()) 										//////
		{ 																//////
			PointerPositionByte = ConsoleGet(); 						//////
			bSendUpdate = TRUE; 										//////
            ConsolePutROMString( (ROM char *)" Received data = " ); 	//////
			PrintChar( PointerPositionByte ); 							//////
            ConsolePutROMString( (ROM char *)"\r\n" ); 					//////
		} 																//////

    }
}


/*******************************************************************************
User Interrupt Handler

The stack uses some interrupts for its internal processing.  Once it is done
checking for its interrupts, the stack calls this function to allow for any
additional interrupt processing.
*******************************************************************************/

void UserInterruptHandler(void)
{

    // *************************************************************************
    // Place any application-specific interrupt processing here
    // *************************************************************************

    // Is this an interrupt-on-change interrupt?
    if ( RBIF == 1 )
    {
        // Record which button was pressed so the main() loop can
        // handle it
//        if (BIND_SWITCH == 0)
//            myStatusFlags.bits.bBindSwitchToggled = TRUE;			////// we don't have a bind switch 

        #ifdef I_AM_SWITCH
//            if (LIGHT_SWITCH == 0)
//                myStatusFlags.bits.bLightSwitchToggled = TRUE;	////// we don't have a light switch 
        #endif

        // Disable further RBIF until we process it
        RBIE = 0;

        // Clear mis-match condition and reset the interrupt flag
        LATB = PORTB;

        RBIF = 0;
    }
}
