/***********************************************************************************

	Stock Clock Pointer.c

	This file is based on RFD.c in the Microchip Stack for ZigBee version 1.0-1.35
	and you should be able to substitute the file directly in the demo application supplied
    with the stack

	All modified lines end in "//////"

	The stack may be downloaded from www.microchip.com

	You may contact the author Richard Hoptroff at rgh@flexipanel.com

************************************************************************************
/*
    Microchip ZigBee Stack

    Demo RFD

    This demonstration shows how a ZigBee RFD can be set up.  This demo allows
    the PICDEM Z Demostration Board to act as either a "Switching Load Controller"
    (e.g. a light) or a "Switching Remote Control" (e.g. a switch) as defined by
    the Home Controls, Lighting profile.  It is designed to interact with a
    second PICDEM Z programmed with the Demo Coordinator project.

    To give the PICDEM Z "switch" capability, uncomment the I_AM_SWITCH definition
    below.  To give the PICDEM Z "light" capability, uncomment the I_AM_LIGHT
    definition below.  The PICDEM Z may have both capabilities enabled.  Be sure
    that the corresponding Demo Coordinator device is programmed with complementary
    capabilities.  NOTE - for simplicity, the ZigBee simple descriptors for this
    demonstration are fixed.

    If this node is configured as a "switch", it can discover the network address
    of the "light" using two methods.  If the USE_BINDINGS definition is
    uncommented below, then End Device Binding must be performed between the
    "switch" and the "light" before messages can be sent and received successfully.
    If USE_BINDINGS is commented out, then the node will default to the probable
    network address of the other node, and messages may be able to be sent
    immediately.  However, the node will also be capable of performing Device
    Discovery to discover the actual network address of the other node, in case
    the network was formed with alternate short address assignments.  NOTE: The
    USE_BINDINGS definition must be the same in both the RFD and the ZigBee
    Coordinator nodes.

    Switch functionality is as follows:
        RB4, I_AM_SWITCH defined, sends a "toggle" message to the other node's "light"
        RB4, I_AM_SWITCH not defined, no effect
        RB5, USE_BINDINGS defined, sends an End Device Bind request
        RB5, USE_BINDINGS undefined, sends a NWK_ADDR_req for the MAC address specified

    End Device Binding
    ------------------
    If the USE_BINDINGS definition is uncommented, the "switch" will send an
    APS indirect message to toggle the "light".  In order for the message to
    reach its final destination, a binding must be created between the "switch"
    and the "light".  To do this, press RB5 on one PICDEM Z, and then press RB5
    on the other PICDEM Z within 5 seconds.  A message will be displayed indicating
    if binding was successful or not.  Note that End Device Binding is a toggle
    function.  Performing the operation again will unbind the nodes, and messages
    will not reach their final destination.

    Device Discovery
    ----------------
    If the USE_BINDINGS definition is not uncommented, pressing RB5 will send a
    broadcast NWK_ADDR_req message.  The NWK_ADDR_req message contains the MAC
    address of the desired node.  Be sure this address matches the address
    contained in the other node's zigbee.def file.

 *********************************************************************
 * FileName:        RFD.c
 * Dependencies:
 * Processor:       PIC18F
 * Complier:        MCC18 v3.00 or higher
 * Company:         Microchip Technology, Inc.
 *
 * Software License Agreement
 *
 * The software supplied herewith by Microchip Technology Incorporated
 * (the Company) for its PICmicro Microcontroller is intended and
 * supplied to you, the Companys customer, for use solely and
 * exclusively on Microchip PICmicro Microcontroller products. The
 * software is owned by the Company and/or its supplier, and is
 * protected under applicable copyright laws. All rights are reserved.
 * Any use in violation of the foregoing restrictions may subject the
 * user to criminal sanctions under applicable laws, as well as to
 * civil liability for the breach of the terms and conditions of this
 * license.
 *
 * THIS SOFTWARE IS PROVIDED IN AN AS IS CONDITION. NO WARRANTIES,
 * WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED
 * TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. THE COMPANY SHALL NOT,
 * IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL OR
 * CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
 *
 *
 * Author               Date    Comment
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * DF/KO                01/09/06 Microchip ZigBee Stack v1.0-3.4.1
 ********************************************************************/


//******************************************************************************
// Header Files
//******************************************************************************

// Include the main ZigBee header file.
#include "zAPL.h"

// If you are going to send data to a terminal, include this file.
#include "console.h"

////// We use the nonvolatile memory routines
#include "zNVM.h"		//////

////// We use the time routines routines
#include "SymbolTime.h"		//////

//******************************************************************************
// Configuration Bits
//******************************************************************************

#define OSC_INTERNAL									////// use the internal oscillator since ztal not fitted as standard on Pixies
//#define WDT_2_MIN										////// 'release' version with 2 minute wakeup
#define WDT_2_SEC										////// 'test' version with 2 second wakeup

#if defined(MCHP_C18) && defined(__18F4620)
    #pragma romdata CONFIG1H = 0x300001
#if defined(OSC_4MHz)
    const rom unsigned char config1H = 0b00000110;      ////// HSPLL oscillator
#elif defined(OSC_16MHz)
    const rom unsigned char config1H = 0b00000010;     	////// HS oscillator
#else
    const rom unsigned char config1H = 0b00001000;      ////// internal oscillator
#endif

    #pragma romdata CONFIG2L = 0x300002
    const rom unsigned char config2L = 0b00011111;      // Brown-out Reset Enabled in hardware @ 2.0V, PWRTEN disabled

   #pragma romdata CONFIG2H = 0x300003
#ifdef WDT_2_MIN										//////
    const rom unsigned char config2H = 0b00011110;      ////// HW WD disabled, 4ms x 1:32768 prescaler = every 2 minutes
#endif
#ifdef WDT_2_SEC										//////
    const rom unsigned char config2H = 0b00010010;      ////// HW WD disabled, 4ms x 1:512 prescaler = every 2 seconds
#endif

    #pragma romdata CONFIG3H = 0x300005
    const rom unsigned char config3H = 0b10000000;      // PORTB digital on RESET

    #pragma romdata CONFIG4L = 0x300006
    const rom unsigned char config4L = 0b10000001;      // DEBUG disabled,
                                                        // XINST disabled
                                                        // LVP disabled
                                                        // STVREN enabled

    #pragma romdata

#elif defined(HITECH_C18) && defined(_18F4620)
    // Set configuration fuses for HITECH compiler.
    __CONFIG(1, 0x0600);    // HSPLL oscillator
    __CONFIG(2, 0x101F);    // PWRTEN disabled, BOR enabled @ 2.0V, HW WD disabled, 1:128 prescaler
    __CONFIG(3, 0x8000);    // PORTB digital on RESET
    __CONFIG(4, 0x0081);    // DEBUG disabled,
                            // XINST disabled
                            // LVP disabled
                            // STVREN enabled
#endif

//******************************************************************************
// Compilation Configuration
//******************************************************************************

#define USE_BINDINGS	//////
#define I_AM_LIGHT		//////  The end device only receives switch information
//#define I_AM_SWITCH	//////  The end device only receives switch information

//******************************************************************************
// Constants
//******************************************************************************

// Switches and LEDs locations.
#define BROADCAST_SWITCH            RB5
#define LIGHT_SWITCH                RB4

#define BIND_INDICATION             LATA0

////// change message indication to RE0 since RA1 used by stepper
#define MESSAGE_INDICATION      	LATEbits.LATE0			//////
#define TRIS_MESSAGE_INDICATION 	TRISEbits.TRISE0		//////

#define BIND_STATE_BOUND            0
#define BIND_STATE_TOGGLE           1
#define BIND_STATE_UNBOUND          1
#define BIND_WAIT_DURATION          (6*ONE_SECOND)

#define LIGHT_OFF                   0x00
#define LIGHT_ON                    0xFF
#define LIGHT_TOGGLE                0xF0


//******************************************************************************
// Application Variables
//******************************************************************************

static union
{
    struct
    {
        BYTE    bBroadcastSwitchToggled    : 1;
        BYTE    bLightSwitchToggled        : 1;
        BYTE    bTryingToBind              : 1;
        BYTE    bIsBound                   : 1;
        BYTE    bDestinationAddressKnown   : 1;
    } bits;
    BYTE Val;
} myStatusFlags;
#define STATUS_FLAGS_INIT       0x00	
#define TOGGLE_BOUND_FLAG       0x08

NETWORK_DESCRIPTOR  *currentNetworkDescriptor;
ZIGBEE_PRIMITIVE    currentPrimitive;
SHORT_ADDR          destinationAddress;
NETWORK_DESCRIPTOR  *NetworkDescriptor;


//******************************************************************************
// Function Prototypes
//******************************************************************************

BOOL myProcessesAreDone( void );

//////******************************************************************************

////// Definitions and declarations
#define IsBound	myStatusFlags.bits.bIsBound					////// Abbreviation to fit the listing in Circuit Cellar
#define bBindSwitchToggled      bBroadcastSwitchToggled		////// Easier to understand if referred to as bind switch

////// ROM data that gets modified must be in a reserved, aligned 0x40 region of memory 
////// to ensure NVMErase / NVMwrite functions do not modify executable code
#pragma romdata NV_MEM = 0xFD40				////// Align to 0x40 boundary, immediately prior to debug section
ROM	BOOL	NonvolatileBindFlag = FALSE;	////// Store bind state for next power-up
ROM	BYTE	NonvolatileFiller[0x3F];		////// fill up remainder of 0x40 section
#pragma romdata								//////

////// motor control

void ClockwiseStep( void );					//////
void CounterClockwiseStep( void );			//////
#include <delays.h>							//////
#define C1_HIGH			0b00000010			//////
#define C23_HIGH		0b00001100			//////
#define C4_HIGH 		0b00010000			//////
#define NOT_C_PINS		0b11100001			//////
#define TRISA_SETTINGS	0b11100001			//////
#define NUDGE_SWITCH	BROADCAST_SWITCH	//////

////// dial position store
static BYTE CurrentDialPosition = 0;						//////
static BYTE DesiredDialPosition = 0;						//////
#define	SetDialPosition( x ) DesiredDialPosition = x;		//////

//******************************************************************************
//******************************************************************************
// Main
//******************************************************************************
//******************************************************************************

void main(void)
{
#ifdef OSC_INTERNAL
	OSCCONbits.IRCF1 = 1;  // assuming <IDCF2:IDCF0>=100 on startup, changes to <IDCF2:IDCF0>=110 = 4MHz
	OSCTUNEbits.PLLEN = 1;  // PLL 4MHz -> 16MHz
#endif

    CLRWDT();
    ENABLE_WDT();

    currentPrimitive = NO_PRIMITIVE;
    NetworkDescriptor = NULL;

    // If you are going to send data to a terminal, initialize the UART.
    ConsoleInit();

    // Initialize the ZigBee Stack.
    ZigBeeInit();

    // *************************************************************************
    // Perform any other initialization here
    // *************************************************************************

    myStatusFlags.Val = STATUS_FLAGS_INIT;

	////// set up I/O for motor and modified MESSAGE_INDICATION
	LATA = (LATA & NOT_C_PINS) ;		//////
	TRISA &= TRISA_SETTINGS;			//////
	TRIS_MESSAGE_INDICATION = 0;		//////

	////// Get bind flag from nonvolatile memory
	IsBound = NonvolatileBindFlag;	//////

	////// if not bound, bind immediately on startup.  Note this make the implementation very easy but to rebind you need
	////// to reload the firmware on both tyhe coordinator and the end device.  In a commercial application, you would 
    ////// want to use a bind button.  If you look at the data sheet for Pixie Switcher (on www.flexipanel.com) you'll 
    ////// get an idea of how to implement things like this
	if (!IsBound) myStatusFlags.bits.bBindSwitchToggled = TRUE;  //////

    // Default the destination address to the ZigBee Coordinator
    destinationAddress.Val = 0x0000;

    // Initialize the LED's.
    BIND_INDICATION = !myStatusFlags.bits.bIsBound;
    MESSAGE_INDICATION = 0;

////// As you can see from the next few lines, this template code from Microchip is actually
////// their beta version, but the release version will be vailable by the time the article is
////// in press, so you would be wise to download the latest version of the Microchip source
////// and modify that.
    ConsolePutROMString( (ROM char *)"\r\n\r\n\r\n********************************\r\n" );
    ConsolePutROMString( (ROM char *)"Microchip ZigBee(TM) Stack - pre-v1.0-4.0!\r\n\r\n" );
    ConsolePutROMString( (ROM char *)"ZigBee RFD - Stock Clock Version\r\n\r\n" );		//////

   while (1)
    {
        CLRWDT();
        ZigBeeTasks( &currentPrimitive );

        switch (currentPrimitive)
        {
            case NLME_NETWORK_DISCOVERY_confirm:
                currentPrimitive = NO_PRIMITIVE;
                if (!params.NLME_NETWORK_DISCOVERY_confirm.Status)
                {
                    if (!params.NLME_NETWORK_DISCOVERY_confirm.NetworkCount)
                    {
                        ConsolePutROMString( (ROM char *)"No networks found.  Trying again...\r\n" );
                    }
                    else
                    {
                        // Save the descriptor list pointer so we can destroy it later.
                        NetworkDescriptor = params.NLME_NETWORK_DISCOVERY_confirm.NetworkDescriptor;

                        // Select a network to try to join.  We're not going to be picky right now...
                        currentNetworkDescriptor = NetworkDescriptor;

                        // not needed for new join params.NLME_JOIN_request.ScanDuration = ;
                        // not needed for new join params.NLME_JOIN_request.ScanChannels = ;
                        params.NLME_JOIN_request.PANId          = currentNetworkDescriptor->PanID;
                        ConsolePutROMString( (ROM char *)"Network(s) found. Trying to join " );
                        PrintChar( params.NLME_JOIN_request.PANId.byte.MSB );
                        PrintChar( params.NLME_JOIN_request.PANId.byte.LSB );
                        ConsolePutROMString( (ROM char *)".\r\n" );
                        params.NLME_JOIN_request.JoinAsRouter   = FALSE;
                        params.NLME_JOIN_request.RejoinNetwork  = FALSE;
                        params.NLME_JOIN_request.PowerSource    = NOT_MAINS_POWERED;
                        params.NLME_JOIN_request.RxOnWhenIdle   = FALSE;
                        params.NLME_JOIN_request.MACSecurity    = FALSE;
                        currentPrimitive = NLME_JOIN_request;
                    }
                }
                else
                {
                    PrintChar( params.NLME_NETWORK_DISCOVERY_confirm.Status );
                    ConsolePutROMString( (ROM char *)" Error finding network.  Trying again...\r\n" );
                }
                break;

            case NLME_JOIN_confirm:
                if (!params.NLME_JOIN_confirm.Status)
                {
                    ConsolePutROMString( (ROM char *)"Join successful!\r\n" );
                    if (NetworkDescriptor)
                    {
                        // If we joined as an orphan, this will be NULL.
                        free( NetworkDescriptor );
                    }
                    currentPrimitive = NO_PRIMITIVE;

					////// Error in beta version of stack, this flag does not get cleared, clear it now
					////// This should be fixed by the time the stack is reelased in March 2006
					{
						extern MAC_TASKS_PENDING macTasksPending;
						macTasksPending.bits.channelScanning = 0;
					}
                }
                else
                {
                    PrintChar( params.NLME_JOIN_confirm.Status );
                    ConsolePutROMString( (ROM char *)" Could not join.\r\n" );
                    //ZigBeeStatus.flags.bits.bJoiningNew = 1;
                    currentPrimitive = NO_PRIMITIVE;
                }
                break;

            case APSDE_DATA_indication:
                {
                    WORD_VAL    attributeId;
                    BYTE        command;
                    BYTE        data;
                    //BYTE        dataType;
                    BYTE        frameHeader;
                    BYTE        sequenceNumber;
                    BYTE        transaction;
                    BYTE        transByte;

                    currentPrimitive = NO_PRIMITIVE;
                    frameHeader = APLGet();

                    switch (params.APSDE_DATA_indication.DstEndpoint)
                    {
                        case EP_ZDO:
                            #define dataLength command

                            ConsolePutROMString( (ROM char *)"  Receiving ZDO cluster " );
                            PrintChar( params.APSDE_DATA_indication.ClusterId );
                            ConsolePutROMString( (ROM char *)"\r\n" );

                            // Put code here to handle any ZDO responses that we requested
                            if ((frameHeader & APL_FRAME_TYPE_MASK) == APL_FRAME_TYPE_MSG)
                            {
                                frameHeader &= APL_FRAME_COUNT_MASK;
                                for (transaction=0; transaction<frameHeader; transaction++)
                                {
                                    sequenceNumber          = APLGet();
                                    dataLength              = APLGet();
                                    transByte               = 1;    // Account for status byte

                                    switch( params.APSDE_DATA_indication.ClusterId )
                                    {

                                        // ********************************************************
                                        // Put a case here to handle each ZDO response that we requested.
                                        // ********************************************************

                                        case NWK_ADDR_rsp:
                                            if (APLGet() == SUCCESS)
                                            {
                                                ConsolePutROMString( (ROM char *)"  Receiving NWK_ADDR_rsp.\r\n" );

                                                // Skip over the IEEE address of the responder.
                                                for (data=0; data<8; data++)
                                                {
                                                    APLGet();
                                                    transByte++;
                                                }
                                                destinationAddress.byte.LSB = APLGet();
                                                destinationAddress.byte.MSB = APLGet();
                                                transByte += 2;
                                                myStatusFlags.bits.bDestinationAddressKnown = 1;
                                            }
                                            break;

                                        #ifdef USE_BINDINGS
                                        case END_DEVICE_BIND_rsp:
                                            switch( APLGet() )
                                            {
                                                case SUCCESS:
                                                    ConsolePutROMString( (ROM char *)" End device bind/unbind successful!\r\n" );
                                                    myStatusFlags.bits.bIsBound = !(myStatusFlags.bits.bIsBound);
                                                    BIND_INDICATION = !myStatusFlags.bits.bIsBound;

													{	//////
														BYTE flag, iCount;													//////

														////// Store new bind state in nonvolatile memory
														flag = myStatusFlags.bits.bIsBound;									//////
														NVMWrite((NVM_ADDR*)&NonvolatileBindFlag, &flag, sizeof(BYTE)); 	//////

														////// rotate pointer 90 degrees to indicate bind success
														CurrentDialPosition = 45;											//////
														DesiredDialPosition = 0;											//////
													}	////// 

                                                    break;
                                                case ZDO_NOT_SUPPORTED:
                                                    ConsolePutROMString( (ROM char *)" End device bind/unbind not supported.\r\n" );
                                                    break;
                                                case END_DEVICE_BIND_TIMEOUT:
                                                    ConsolePutROMString( (ROM char *)" End device bind/unbind time out.\r\n" );
                                                    break;
                                                case END_DEVICE_BIND_NO_MATCH:
                                                    ConsolePutROMString( (ROM char *)" End device bind/unbind failed - no match.\r\n" );
                                                    break;
                                                default:
                                                    ConsolePutROMString( (ROM char *)" End device bind/unbind invalid response.\r\n" );
                                                    break;
                                            }
                                            myStatusFlags.bits.bTryingToBind = 0;
                                            break;
                                        #endif

                                        default:
                                            break;
                                    }

                                    // Read out the rest of the MSG in case there is another transaction.
                                    for (; transByte<dataLength; transByte++)
                                    {
                                        APLGet();
                                    }
                                }
                            }
                            #undef dataLength
                            break;

                        // ************************************************************************
                        // Place a case for each user defined endpoint.
                        // ************************************************************************

                        case EP_LIGHT:
                            if ((frameHeader & APL_FRAME_TYPE_MASK) == APL_FRAME_TYPE_KVP)
                            {
                                frameHeader &= APL_FRAME_COUNT_MASK;
                                for (transaction=0; transaction<frameHeader; transaction++)
                                {
                                    sequenceNumber          = APLGet();
                                    command                 = APLGet();
                                    attributeId.byte.LSB    = APLGet();
                                    attributeId.byte.MSB    = APLGet();

                                    //dataType = command & APL_FRAME_DATA_TYPE_MASK;
                                    command &= APL_FRAME_COMMAND_MASK;

                                    if ((params.APSDE_DATA_indication.ClusterId == OnOffSRC_CLUSTER) &&
                                        (attributeId.Val == OnOffSRC_OnOff))
                                    {
                                        if ((command == APL_FRAME_COMMAND_SET) ||
                                            (command == APL_FRAME_COMMAND_SETACK))
                                        {
                                            // Prepare a response in case it is needed.
                                            TxBuffer[TxData++] = APL_FRAME_TYPE_KVP | 1;    // KVP, 1 transaction
                                            TxBuffer[TxData++] = sequenceNumber;
//                                            TxBuffer[TxData++] = APL_FRAME_COMMAND_SET_RES | TRANS_UINT8 ;	////// Error in Microchip demo application
                                            TxBuffer[TxData++] = APL_FRAME_COMMAND_SET_RES | (TRANS_UINT8 << 4);	////// TRANS_UINT8 must be left-shifted by 4
                                            TxBuffer[TxData++] = attributeId.byte.LSB;
                                            TxBuffer[TxData++] = attributeId.byte.MSB;
 
											////// Set dial position
											SetDialPosition( APLGet() );								//////					
				 				            ConsolePutROMString( (ROM char *)" Received data = " ); 	//////
											PrintChar( DesiredDialPosition ); 							//////
								            ConsolePutROMString( (ROM char *)"\r\n" ); 					//////

/*											////// original demo code commented out
                                            // Data type for this attibute must be TRANS_UINT8
                                            data = APLGet();
                                            switch (data)
                                            {
                                                case LIGHT_OFF:
                                                    ConsolePutROMString( (ROM char *)" Turning light off.\r\n" );
                                                    MESSAGE_INDICATION = 0;
                                                    TxBuffer[TxData++] = SUCCESS;
                                                    break;
                                                case LIGHT_ON:
                                                    ConsolePutROMString( (ROM char *)" Turning light on.\r\n" );
                                                    MESSAGE_INDICATION = 1;
                                                    TxBuffer[TxData++] = SUCCESS;
                                                    break;
                                                case LIGHT_TOGGLE:
                                                    ConsolePutROMString( (ROM char *)" Toggling light.\r\n" );
                                                    MESSAGE_INDICATION ^= 1;
                                                    TxBuffer[TxData++] = SUCCESS;
                                                    break;
                                                default:
                                                    PrintChar( data );
                                                    ConsolePutROMString( (ROM char *)" Invalid light message.\r\n" );
                                                    TxBuffer[TxData++] = KVP_INVALID_ATTRIBUTE_DATA;
                                                    break;
                                            }
 */
                                       }
                                        if (command == APL_FRAME_COMMAND_SETACK)
                                        {
                                            // Send back an application level acknowledge.
                                            ZigBeeBlockTx();

                                            // Take care here that parameters are not overwritten before they are used.
                                            // We can use the data byte as a temporary variable.
                                            params.APSDE_DATA_request.DstAddrMode = params.APSDE_DATA_indication.SrcAddrMode;
                                            params.APSDE_DATA_request.DstEndpoint = params.APSDE_DATA_indication.SrcEndpoint;
                                            params.APSDE_DATA_request.DstAddress.ShortAddr = params.APSDE_DATA_indication.SrcAddress.ShortAddr;

                                            //params.APSDE_DATA_request.asduLength; TxData
                                            //params.APSDE_DATA_request.ProfileId; unchanged
                                            params.APSDE_DATA_request.RadiusCounter = DEFAULT_RADIUS;
                                            params.APSDE_DATA_request.DiscoverRoute = ROUTE_DISCOVERY_ENABLE;
                                            params.APSDE_DATA_request.TxOptions.Val = 0;
                                            params.APSDE_DATA_request.SrcEndpoint = EP_LIGHT;
                                            //params.APSDE_DATA_request.ClusterId; unchanged

                                            currentPrimitive = APSDE_DATA_request;
                                        }
                                        else
                                        {
                                            // We are not sending an acknowledge, so reset the transmit message pointer.
                                            TxData = TX_DATA_START;
                                        }
                                    }
                                    // TODO read to the end of the transaction.
                                } // each transaction
                            } // frame type
                            break;
                        default:
                            // If the command type was something that requested an acknowledge, we could send back
                            // KVP_INVALID_ENDPOINT here.
                            break;
                    }

                    APLDiscardRx();
                }
                break;

            case APSDE_DATA_confirm:
                if (params.APSDE_DATA_confirm.Status)
                {
                    ConsolePutROMString( (ROM char *)"Error " );
                    PrintChar( params.APSDE_DATA_confirm.Status );
                    ConsolePutROMString( (ROM char *)" sending message.\r\n" );
                }
                else
                {
                    ConsolePutROMString( (ROM char *)" Message sent successfully.\r\n" );
                }
                currentPrimitive = NO_PRIMITIVE;
                break;

            case NLME_SYNC_confirm:
                switch (params.NLME_SYNC_confirm.Status)
                {
                    case NWK_SYNC_FAILURE:
                        // We don't know if our parent is gone or simply has no data...
                        ConsolePutROMString( (ROM char *)"Data request complete.\r\n" );
                        break;

                    case NWK_INVALID_PARAMETER:
                        // If we call NLME_SYNC_request correctly, this doesn't occur.
                        ConsolePutROMString( (ROM char *)"Invalid sync parameter.\r\n" );
                        break;

                    case SUCCESS:
                        // I have data.  I can wait for the Data Indication
                        // NOTE: in the Microchip stack, this is not returned.  We just
                        // receive the NLME_DATA_request.
                        ConsolePutROMString( (ROM char *)"Data is coming for us...\r\n" );
                        break;
                }
                currentPrimitive = NO_PRIMITIVE;
                break;

            case NO_PRIMITIVE:
                if (!ZigBeeStatus.flags.bits.bNetworkJoined)
                {
                    if (!ZigBeeStatus.flags.bits.bTryingToJoinNetwork)
                    {
                        if (ZigBeeStatus.flags.bits.bTryOrphanJoin)
                        {
                            ConsolePutROMString( (ROM char *)"Trying to join network as an orphan...\r\n" );
                            params.NLME_JOIN_request.JoinAsRouter           = FALSE;
                            params.NLME_JOIN_request.RejoinNetwork          = TRUE;
                            params.NLME_JOIN_request.PowerSource            = NOT_MAINS_POWERED;
                            params.NLME_JOIN_request.RxOnWhenIdle           = FALSE;
                            params.NLME_JOIN_request.MACSecurity            = FALSE;
                            params.NLME_JOIN_request.ScanDuration           = 8;
                            params.NLME_JOIN_request.ScanChannels.Val       = ALLOWED_CHANNELS;
                            currentPrimitive = NLME_JOIN_request;
                        }
                        else
                        {
                            ConsolePutROMString( (ROM char *)"Trying to join network as a new device...\r\n" );
                            params.NLME_NETWORK_DISCOVERY_request.ScanDuration          = 8;
                            params.NLME_NETWORK_DISCOVERY_request.ScanChannels.Val      = ALLOWED_CHANNELS;
                            currentPrimitive = NLME_NETWORK_DISCOVERY_request;
                        }
                    }
                }
                else
                {
                    // See if I can do my own internal tasks.  We don't want to try to send a message
                    // if we just asked for one.
                    if (ZigBeeStatus.flags.bits.bDataRequestComplete & ZigBeeReady())
                    {

                        // ************************************************************************
                        // Place all processes that can send messages here.  Be sure to call
                        // ZigBeeBlockTx() when currentPrimitive is set to APSDE_DATA_request.
                        // ************************************************************************

                        if ( myStatusFlags.bits.bLightSwitchToggled )
                        {
                            // Send a light toggle message to the other node.
                            myStatusFlags.bits.bLightSwitchToggled = FALSE;

                            TxBuffer[TxData++] = APL_FRAME_TYPE_KVP | 1;    // KVP, 1 transaction
                            TxBuffer[TxData++] = APLGetTransId();
//                            TxBuffer[TxData++] = APL_FRAME_COMMAND_SET | TRANS_UINT8;				////// Error in Microchip demo application, will probably be fixed
                            TxBuffer[TxData++] = APL_FRAME_COMMAND_SET | (TRANS_UINT8 << 4);		////// by the time CC article is in print TRANS_UINT8 must be left-shifted by 4
                            TxBuffer[TxData++] = OnOffSRC_OnOff & 0xFF;         // Attribute ID LSB
                            TxBuffer[TxData++] = (OnOffSRC_OnOff >> 8) & 0xFF;  // Attribute ID MSB
                            TxBuffer[TxData++] = LIGHT_TOGGLE;

                            #ifdef USE_BINDINGS
                                params.APSDE_DATA_request.DstAddrMode = APS_ADDRESS_NOT_PRESENT;
                            #else
                                params.APSDE_DATA_request.DstAddrMode = APS_ADDRESS_16_BIT;
                                params.APSDE_DATA_request.DstEndpoint = EP_LIGHT;
                                params.APSDE_DATA_request.DstAddress.ShortAddr = destinationAddress;
                            #endif

                            //params.APSDE_DATA_request.asduLength; TxData
                            params.APSDE_DATA_request.ProfileId.Val = MY_PROFILE_ID;
                            params.APSDE_DATA_request.RadiusCounter = DEFAULT_RADIUS;
                            params.APSDE_DATA_request.DiscoverRoute = ROUTE_DISCOVERY_ENABLE;
                            params.APSDE_DATA_request.TxOptions.Val = 0;
                            params.APSDE_DATA_request.SrcEndpoint = EP_LIGHT;
                            params.APSDE_DATA_request.ClusterId = OnOffSRC_CLUSTER;

                            ConsolePutROMString( (ROM char *)" Trying to send light switch message.\r\n" );

                            currentPrimitive = APSDE_DATA_request;
                        }
                        else
                        #ifdef USE_BINDINGS
                            if (myStatusFlags.bits.bBroadcastSwitchToggled)
                            {
                                // Send END_DEVICE_BIND_req

                                myStatusFlags.bits.bBroadcastSwitchToggled = FALSE;

                                TxBuffer[TxData++] = APL_FRAME_TYPE_MSG | 1;    // KVP, 1 transaction
                                TxBuffer[TxData++] = APLGetTransId();
                                #if defined(I_AM_LIGHT) && defined(I_AM_SWITCH)
                                    TxBuffer[TxData++] = 9; // Transaction Length
                                #else
                                    TxBuffer[TxData++] = 8; // Transaction Length
                                #endif

                                // Binding Target
                                TxBuffer[TxData++] = 0x00;      // Binding Target
                                TxBuffer[TxData++] = 0x00;
                                TxBuffer[TxData++] = EP_LIGHT;
                                TxBuffer[TxData++] = MY_PROFILE_ID_LSB;
                                TxBuffer[TxData++] = MY_PROFILE_ID_MSB;

                                #ifdef I_AM_LIGHT
                                    TxBuffer[TxData++] = 1;     // Input clusters
                                    TxBuffer[TxData++] = OnOffSRC_CLUSTER;
                                #else
                                    TxBuffer[TxData++] = 0;
                                #endif

                                #ifdef I_AM_SWITCH
                                    TxBuffer[TxData++] = 1;     // Output clusters
                                    TxBuffer[TxData++] = OnOffSRC_CLUSTER;
                                #else
                                    TxBuffer[TxData++] = 0;
                                #endif

                                params.APSDE_DATA_request.DstAddrMode = APS_ADDRESS_16_BIT;
                                params.APSDE_DATA_request.DstEndpoint = EP_ZDO;
                                params.APSDE_DATA_request.DstAddress.ShortAddr.Val = 0x0000;

                                //params.APSDE_DATA_request.asduLength; TxData
                                params.APSDE_DATA_request.ProfileId.Val = MY_PROFILE_ID;
                                params.APSDE_DATA_request.RadiusCounter = DEFAULT_RADIUS;
                                params.APSDE_DATA_request.DiscoverRoute = ROUTE_DISCOVERY_ENABLE;
                                params.APSDE_DATA_request.TxOptions.Val = 0;
                                params.APSDE_DATA_request.SrcEndpoint = EP_ZDO;
                                params.APSDE_DATA_request.ClusterId = END_DEVICE_BIND_req;

                                ConsolePutROMString( (ROM char *)" Trying to send END_DEVICE_BIND_req.\r\n" );

                                currentPrimitive = APSDE_DATA_request;
                            }
                        #else
                            if (myStatusFlags.bits.bBroadcastSwitchToggled)
                            {
                                // Send NWK_ADDR_req

                                myStatusFlags.bits.bBroadcastSwitchToggled = FALSE;

                                TxBuffer[TxData++] = APL_FRAME_TYPE_MSG | 1;    // KVP, 1 transaction
                                TxBuffer[TxData++] = APLGetTransId();
                                TxBuffer[TxData++] = 10; // Transaction Length

                                // IEEEAddr of the node we want to find.  !!!Must match our other PICDEM Z!!!
                                TxBuffer[TxData++] = 0x54;
                                TxBuffer[TxData++] = 0x00;
                                TxBuffer[TxData++] = 0x00;
                                TxBuffer[TxData++] = 0x00;
                                TxBuffer[TxData++] = 0x00;
                                TxBuffer[TxData++] = 0xa3;
                                TxBuffer[TxData++] = 0x04;
                                TxBuffer[TxData++] = 0x00;

                                // RequestType
                                TxBuffer[TxData++] = 0x00;

                                // StartIndex
                                TxBuffer[TxData++] = 0x00;

                                params.APSDE_DATA_request.DstAddrMode = APS_ADDRESS_16_BIT;
                                params.APSDE_DATA_request.DstEndpoint = EP_ZDO;
                                params.APSDE_DATA_request.DstAddress.ShortAddr.Val = 0xFFFF;

                                //params.APSDE_DATA_request.asduLength; TxData
                                params.APSDE_DATA_request.ProfileId.Val = MY_PROFILE_ID;
                                params.APSDE_DATA_request.RadiusCounter = DEFAULT_RADIUS;
                                params.APSDE_DATA_request.DiscoverRoute = ROUTE_DISCOVERY_ENABLE;
                                params.APSDE_DATA_request.TxOptions.Val = 0;
                                params.APSDE_DATA_request.SrcEndpoint = EP_ZDO;
                                params.APSDE_DATA_request.ClusterId = NWK_ADDR_req;

                                ConsolePutROMString( (ROM char *)" Trying to send NWK_ADDR_req.\r\n" );

                                currentPrimitive = APSDE_DATA_request;
                            }
                        #endif
                    }

                    // Try to go to sleep.  If we sleep, request data when we wake up.
                    if ((currentPrimitive == NO_PRIMITIVE) && !ZigBeeStatus.flags.bits.bHasBackgroundTasks &&
                         myProcessesAreDone() && ZigBeeStatus.flags.bits.bDataRequestComplete )
                    {
                        // Go to sleep.  Make sure that the UART is finished, turn off the transceiver,
                        // and make sure that we wakeup from key press.
                        while (!ConsoleIsPutReady());
                        APLDisable();
                        RBIE = 1;
                        SLEEP();
                        NOP();

                        // Turn on the transceiver.

                        APLEnable();

                        // Request data from our parent.
                        params.NLME_SYNC_request.Track = FALSE;
                        currentPrimitive = NLME_SYNC_request;
                        ConsolePutROMString( (ROM char *)"Requesting data after sleep...\r\n" );
                    }

					////// We must still poll the coordinator for messages if we haven't been sleeping
					else if (currentPrimitive == NO_PRIMITIVE)									//////
					{																			//////
						static TICK LastWakeTick = 0;											//////
						TICK ThisTick = TickGet();												//////
						if ( ThisTick.Val > LastWakeTick.Val + (ONE_SECOND/2) )					//////
						{																		//////
							LastWakeTick.Val = ThisTick.Val	;									//////
	                        // Request data from our parent.									//////
	                        params.NLME_SYNC_request.Track = FALSE;								//////
	                        currentPrimitive = NLME_SYNC_request;								//////
	                        ConsolePutROMString( (ROM char *)"Requesting data...\r\n" );		//////
						}																		//////
					}																			//////	

                }
                break;

            default:
                PrintChar( currentPrimitive );
                ConsolePutROMString( (ROM char *)" Unhandled primitive.\r\n" );
                currentPrimitive = NO_PRIMITIVE;
        }

        // *********************************************************************
        // Place any non-ZigBee related processing here.  Be sure that the code
        // will loop back and execute ZigBeeTasks() in a timely manner.
        // *********************************************************************

		// since my tasks involve calling slow [Counter]ClockwiseStep() functions, only execute once per
		// 250ms so stack gets plenty of opportunity to perform other tasks 

		{																			//////
			////// Has it been 250 ms since the last task?							//////
			static TICK LastNudgeTick = 0;											//////
			TICK ThisTick = TickGet();												//////
			if ( ThisTick.Val > LastNudgeTick.Val + (ONE_SECOND/4) )				//////
			{																		//////
				LastNudgeTick.Val = ThisTick.Val;									//////
	
				////// Is current dial position not the actual dial position?					//////
				////// If so, move dial															//////
				if (CurrentDialPosition > DesiredDialPosition)									//////
				{																				//////
					CurrentDialPosition --;														//////
					CounterClockwiseStep();														//////
	             	ConsolePutROMString( (ROM char *)"Move pointer counterclockwise\r\n" );		//////
				}																				//////
				if (CurrentDialPosition < DesiredDialPosition)									//////
				{																				//////
					CurrentDialPosition ++;														//////
					ClockwiseStep();															//////
	             	ConsolePutROMString( (ROM char *)"Move pointer clockwise\r\n" );			//////
				}																				//////
	
				////// if nudge switch pressed, inch pointer forward 
				if (NUDGE_SWITCH==0)												//////
				{																	//////
					// nudge														//////
					ClockwiseStep();												//////
	             	ConsolePutROMString( (ROM char *)"Nudge\r\n" );					//////
				}																	//////
			}																		//////
		}																			//////

    }
}

/*******************************************************************************
myProcessesAreDone

This routine should contain any tests that are required by the application to
confirm that it can go to sleep.  If the application can go to sleep, this
routine should return TRUE.  If the application is still busy, this routine
should return FALSE.
*******************************************************************************/

BOOL myProcessesAreDone( void )
{
	////// if dial not correct, device should not sleep yet
	if (CurrentDialPosition != DesiredDialPosition)	return FALSE;					//////

	////// if nudging, don't sleep
 	if (NUDGE_SWITCH==0) return FALSE;												//////
	
	////// if binding, don't sleep
	if (!IsBound) return FALSE;														//////

    return (myStatusFlags.bits.bBroadcastSwitchToggled==FALSE) && (myStatusFlags.bits.bLightSwitchToggled==FALSE);
}

/*******************************************************************************
User Interrupt Handler

The stack uses some interrupts for its internal processing.  Once it is done
checking for its interrupts, the stack calls this function to allow for any
additional interrupt processing.
*******************************************************************************/

void UserInterruptHandler(void)
{

    // *************************************************************************
    // Place any application-specific interrupt processing here
    // *************************************************************************

    // Is this a interrupt-on-change interrupt?
    if ( RBIF == 1 )
    {
        // Record which button was pressed so the main() loop can
        // handle it
//        if (BROADCAST_SWITCH == 0)									//////	Commented out since no longer required
//            myStatusFlags.bits.bBroadcastSwitchToggled = TRUE;		//////

//      if (LIGHT_SWITCH == 0)											//////	Commented out since no longer required
//            myStatusFlags.bits.bLightSwitchToggled = TRUE;			//////

        // Disable further RBIF until we process it
        RBIE = 0;

        // Clear mis-match condition and reset the interrupt flag
        LATB = PORTB;

        RBIF = 0;
    }
}

#ifndef NO_MOTOR_CONNECTED
// slow step value to ensure no slippage
#define STEP_DELAY_MS	30	/* assumes 16MHz osc */
void CounterClockwiseStep( void )
{
    CLRWDT();
	// energise in last known coil position
	LATA = (LATA & NOT_C_PINS) | (C1_HIGH + C4_HIGH);
	Delay1KTCYx(STEP_DELAY_MS*4);
    CLRWDT();
	
	// cycle forward through positions
	LATA = (LATA & NOT_C_PINS) | C1_HIGH;
	Delay1KTCYx(STEP_DELAY_MS*4);
    CLRWDT();

	LATA = (LATA & NOT_C_PINS) | (C1_HIGH + C23_HIGH);
	Delay1KTCYx(STEP_DELAY_MS*4);
    CLRWDT();

	LATA = (LATA & NOT_C_PINS) | C23_HIGH;
	Delay1KTCYx(STEP_DELAY_MS*4);
    CLRWDT();

	LATA = (LATA & NOT_C_PINS) | (C23_HIGH + C4_HIGH);
	Delay1KTCYx(STEP_DELAY_MS*4);
    CLRWDT();

	LATA = (LATA & NOT_C_PINS) | C4_HIGH;
	Delay1KTCYx(STEP_DELAY_MS*4);
    CLRWDT();

	LATA = (LATA & NOT_C_PINS) | (C1_HIGH + C4_HIGH);
	Delay1KTCYx(STEP_DELAY_MS*4);
    CLRWDT();

	// de-energise
	LATA = (LATA & NOT_C_PINS) ;

}

void ClockwiseStep( void )
{
    CLRWDT();
	// energise in last known coil position
	LATA = (LATA & NOT_C_PINS) | (C1_HIGH + C4_HIGH);
	Delay1KTCYx(STEP_DELAY_MS*4);
    CLRWDT();
	
	// cycle backware through positions
	LATA = (LATA & NOT_C_PINS) | C4_HIGH;
	Delay1KTCYx(STEP_DELAY_MS*4);
    CLRWDT();

	LATA = (LATA & NOT_C_PINS) | (C23_HIGH + C4_HIGH);
	Delay1KTCYx(STEP_DELAY_MS*4);
    CLRWDT();

	LATA = (LATA & NOT_C_PINS) | C23_HIGH;
	Delay1KTCYx(STEP_DELAY_MS*4);
    CLRWDT();

	LATA = (LATA & NOT_C_PINS) | (C1_HIGH + C23_HIGH);
	Delay1KTCYx(STEP_DELAY_MS*4);
    CLRWDT();

	LATA = (LATA & NOT_C_PINS) | C1_HIGH;
	Delay1KTCYx(STEP_DELAY_MS*4);
    CLRWDT();

	LATA = (LATA & NOT_C_PINS) | (C1_HIGH + C4_HIGH);
	Delay1KTCYx(STEP_DELAY_MS*4);
    CLRWDT();

	// de-energise
	LATA = (LATA & NOT_C_PINS) ;

}

#endif
